#
# PySNMP MIB module MBG-SNMP-ROOT-MIB (http://snmplabs.com/pysmi)
# ASN.1 source file://./MBG-SNMP-ROOT-MIB.mib
# Produced by pysmi-0.3.4 at Fri May  1 22:39:55 2020
# On host grumpy platform Linux version 4.15.0-88-generic by user bjk49
# Using Python version 3.6.9 (default, Apr 18 2020, 01:56:04)
#
OctetString, Integer, ObjectIdentifier = mibBuilder.importSymbols("ASN1", "OctetString", "Integer", "ObjectIdentifier")
NamedValues, = mibBuilder.importSymbols("ASN1-ENUMERATION", "NamedValues")
ConstraintsUnion, ValueSizeConstraint, SingleValueConstraint, ValueRangeConstraint, ConstraintsIntersection = mibBuilder.importSymbols("ASN1-REFINEMENT", "ConstraintsUnion", "ValueSizeConstraint", "SingleValueConstraint", "ValueRangeConstraint", "ConstraintsIntersection")
ModuleCompliance, NotificationGroup = mibBuilder.importSymbols("SNMPv2-CONF", "ModuleCompliance", "NotificationGroup")
IpAddress, Gauge32, ModuleIdentity, Unsigned32, enterprises, Bits, Integer32, ObjectIdentity, iso, Counter64, NotificationType, TimeTicks, Counter32, MibScalar, MibTable, MibTableRow, MibTableColumn, MibIdentifier = mibBuilder.importSymbols("SNMPv2-SMI", "IpAddress", "Gauge32", "ModuleIdentity", "Unsigned32", "enterprises", "Bits", "Integer32", "ObjectIdentity", "iso", "Counter64", "NotificationType", "TimeTicks", "Counter32", "MibScalar", "MibTable", "MibTableRow", "MibTableColumn", "MibIdentifier")
TextualConvention, DisplayString = mibBuilder.importSymbols("SNMPv2-TC", "TextualConvention", "DisplayString")
mbgSnmpRoot = ModuleIdentity((1, 3, 6, 1, 4, 1, 5597))
mbgSnmpRoot.setRevisions(('2012-01-25 07:45', '2011-10-14 06:30',))
if mibBuilder.loadTexts: mbgSnmpRoot.setLastUpdated('201201250745Z')
if mibBuilder.loadTexts: mbgSnmpRoot.setOrganization('Meinberg Radio Clocks GmbH & Co. KG')
class MeinbergSwitch(TextualConvention, Integer32):
    status = 'current'
    subtypeSpec = Integer32.subtypeSpec + ConstraintsUnion(SingleValueConstraint(0, 1))
    namedValues = NamedValues(("off", 0), ("on", 1))

mibBuilder.exportSymbols("MBG-SNMP-ROOT-MIB", mbgSnmpRoot=mbgSnmpRoot, MeinbergSwitch=MeinbergSwitch, PYSNMP_MODULE_ID=mbgSnmpRoot)



import os

import txaio
from pysnmp.hlapi.twisted import (CommunityData, ContextData, ObjectIdentity,
                                  ObjectType, SnmpEngine, UdpTransportTarget,
                                  UsmUserData, getCmd, setCmd)

from socs import mibs

# For logging
txaio.use_twisted()


# https://pysnmp.readthedocs.io/en/latest/faq/pass-custom-mib-to-manager.html
MIB_SOURCE = f"file://{os.path.dirname(mibs.__file__)}"


class SNMPTwister:
    """Helper class for handling SNMP communication with twisted.

    More information can be found in the pySNMP documentation: `PySNMP Examples`_

    .. _PySNMP Examples: https://snmplabs.thola.io/pysnmp/examples/contents.html

    Parameters
    ----------
    address : str
        Address of the SNMP Agent to send GET/SET requests to
    port : int
        Associated port for SNMP communication. Default is 161

    Attributes
    ----------
    snmp_engine : pysnmp.entity.engine.SnmpEngine
        PySNMP engine
    address : str
        Address of the SNMP Agent to send GET/SET requests to
    udp_transport : pysnmp.hlapi.twisted.transport.UdpTransportTarget
        UDP transport for UDP over IPv4
    log : txaio.tx.Logger
        txaio logger object

    """

    def __init__(self, address, port=161):
        self.snmp_engine = SnmpEngine()
        self.address = address
        self.udp_transport = UdpTransportTarget((address, port))
        self.log = txaio.make_logger()

    def _success(self, args):
        """Success callback.

        Taken from Twisted example for SNMPv1 from pySNMP documentation:
        https://snmplabs.thola.io/pysnmp/examples/hlapi/twisted/contents.html

        Returns
        -------
        list
            A sequence of ObjectType class instances representing MIB variables
            returned in SNMP response.

        """
        (error_status, error_index, var_binds) = args

        if error_status:
            self.log.error('%s: %s at %s' % (self.address,
                                             error_status.prettyPrint(),
                                             error_index
                                             and var_binds[int(error_index) - 1][0] or '?'))
        else:
            for var in var_binds:
                self.log.debug(' = '.join([x.prettyPrint() for x in var]))

        return var_binds

    def _failure(self, error_indication):
        """Failure Errback.

        Taken from Twisted example for SNMPv1 from pySNMP documentation:
        https://snmplabs.thola.io/pysnmp/examples/hlapi/twisted/contents.html

        """
        self.log.error('%s failure: %s' % (self.address, error_indication))

    def get(self, oid_list, version):
        """Issue a getCmd to get SNMP OID states.

        Example
        -------
        >>> snmp = SNMPTwister('localhost', 161)
        >>> snmp.get([ObjectType(ObjectIdentity('MBG-SNMP-LTNG-MIB',
                                                'mbgLtNgRefclockState',
                                                1)),
                      ObjectType(ObjectIdentity('MBG-SNMP-LTNG-MIB',
                                                'mbgLtNgRefclockLeapSecondDate',
                                                1))])

        >>> snmp = SNMPTwister('localhost', 161)
        >>> result = snmp.get([('MBG-SNMP-LTNG-MIB', 'mbgLtNgRefclockState', 1),
                               ('MBG-SNMP-LTNG-MIB', 'mbgLtNgRefclockLeapSecondDate', 1)])
        >>> # Simply printing the returned object shows a nice string
        >>> print(result[0])
        MBG-SNMP-LTNG-MIB::mbgLtNgRefclockState.1 = notSynchronized
        >>> # The corresponding integer value is hidden within the returned object
        >>> print(result[0][1]._value)
        2

        Parameters
        ----------
        oid_list : list
            List of high-level MIB Object OIDs. The list elements should either be
            ObjectType, or tuples which define the OIDs, as shown in the
            example above. See `Specifying MIB object`_ for more info.

            .. _Specifying MIB Object:
               https://snmplabs.thola.io/pysnmp/docs/pysnmp-hlapi-tutorial.html#specifying-mib-object
        version : int
            SNMP version for communicaton (1, 2, or 3). All versions supported
            here without auth or privacy. If using v3 the configured username
            on the SNMP device should be 'ocs'. For details on version
            implementation in pysnmp see `SNMP Versions`_.

            .. _SNMP Versions:
               https://snmplabs.thola.io/pysnmp/examples/hlapi/asyncore/sync/manager/cmdgen/snmp-versions.html

        Returns
        ------
        twisted.internet.defer.Deferred
            A Deferred which will callback with the var_binds list from
            self._success. If successful, this will contain a list of ObjectType class
            instances representing MIB variables returned in SNMP response.

        """
        oid_list = [ObjectType(ObjectIdentity(*x).addAsn1MibSource(MIB_SOURCE))
                    if isinstance(x, tuple)
                    else x
                    for x
                    in oid_list]

        if version == 1:
            version_object = CommunityData('public', mpModel=0)  # SNMPv1
        elif version == 2:
            version_object = CommunityData('public')  # SNMPv2c
        elif version == 3:
            version_object = UsmUserData('ocs')  # SNMPv3 (no auth, no privacy)
        else:
            raise ValueError(f'SNMP version {version} not supported.')

        datagram = getCmd(self.snmp_engine,
                          version_object,
                          self.udp_transport,
                          ContextData(),
                          *oid_list)

        datagram.addCallback(self._success).addErrback(self._failure)

        return datagram

    def set(self, oid_list, version, setvalue, community_name='private'):
        """Issue a setCmd to set SNMP OID states.
        See `Modifying MIB variables`_ for more info on setting OID states.

        .. _Modifying MIB variables:
           https://snmplabs.thola.io/pysnmp/examples/hlapi/asyncore/sync/manager/cmdgen/modifying-variables.html

        Parameters
        ----------
        oid_list : list
            List of high-level MIB Object OIDs. The list elements should either be
            ObjectType, or tuples which define the OIDs.
        version : int
            SNMP version for communicaton (1, 2, or 3). All versions supported
            here without auth or privacy. If using v3 the configured username
            on the SNMP device should be 'ocs'.
        setvalue : int
            Integer to set OID. For example, 0 is off and 1 is on for outletControl on the iBootPDU.

        Returns
        ------
        twisted.internet.defer.Deferred
            A Deferred which will callback with the var_binds list from
            self._success. If successful, this will contain a list of ObjectType class
            instances representing MIB variables returned in SNMP response.

        """
        oid_list = [ObjectType(ObjectIdentity(*x).addAsn1MibSource(MIB_SOURCE), setvalue)
                    if isinstance(x, tuple)
                    else x
                    for x
                    in oid_list]

        if version == 1:
            version_object = CommunityData(community_name, mpModel=0)  # SNMPv1
        elif version == 2:
            version_object = CommunityData(community_name)  # SNMPv2c
        elif version == 3:
            version_object = UsmUserData('ocs')  # SNMPv3 (no auth, no privacy)
        else:
            raise ValueError(f'SNMP version {version} not supported.')

        datagram = setCmd(self.snmp_engine,
                          version_object,
                          self.udp_transport,
                          ContextData(),
                          *oid_list)

        datagram.addCallback(self._success).addErrback(self._failure)

        return datagram



import argparse
import socket
import time

from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock

from socs.agents.scpi_psu.drivers import PsuInterface


class ScpiPsuAgent:
    def __init__(self, agent, ip_address, gpib_slot):
        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()

        self.job = None
        self.ip_address = ip_address
        self.gpib_slot = gpib_slot
        self.monitor = False

        self.psu = None

        # Registers Temperature and Voltage feeds
        agg_params = {
            'frame_length': 10 * 60,
        }
        self.agent.register_feed('psu_output',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=0)

    @ocs_agent.param('_')
    def init(self, session, params=None):
        """init()

        **Task** - Initialize connection to the power supply.

        """
        with self.lock.acquire_timeout(0) as acquired:
            if not acquired:
                return False, "Could not acquire lock"

            try:
                self.psu = PsuInterface(self.ip_address, self.gpib_slot)
                self.idn = self.psu.identify()
            except socket.timeout as e:
                self.log.error(f"PSU timed out during connect: {e}")
                return False, "Timeout"
            self.log.info("Connected to psu: {}".format(self.idn))

        return True, 'Initialized PSU.'

    @ocs_agent.param('wait', type=float, default=1)
    @ocs_agent.param('channels', type=list, default=[1, 2, 3])
    @ocs_agent.param('test_mode', type=bool, default=False)
    def monitor_output(self, session, params=None):
        """monitor_output(wait=1, channels=[1, 2, 3], test_mode=False)

        **Process** - Continuously monitor PSU output current and voltage.

        Parameters:
            wait (float, optional): Time to wait between measurements
                [seconds].
            channels (list[int], optional): Channels to monitor. [1, 2, 3] by
                default.
            test_mode (bool, optional): Exit process after single loop if True.
                Defaults to False.

        """
        session.set_status('running')
        self.monitor = True

        while self.monitor:
            with self.lock.acquire_timeout(1) as acquired:
                if acquired:
                    data = {
                        'timestamp': time.time(),
                        'block_name': 'output',
                        'data': {}
                    }

                    for chan in params['channels']:
                        data['data']["Voltage_{}".format(chan)] = self.psu.get_volt(chan)
                        data['data']["Current_{}".format(chan)] = self.psu.get_curr(chan)

                    # self.log.info(str(data))
                    # print(data)
                    self.agent.publish_to_feed('psu_output', data)

                    # Allow this process to be queried to return current data
                    session.data = data

                else:
                    self.log.warn("Could not acquire in monitor_current")

            time.sleep(params['wait'])

            if params['test_mode']:
                break

        return True, "Finished monitoring current"

    def stop_monitoring(self, session, params=None):
        self.monitor = False
        return True, "Stopping current monitor"

    @ocs_agent.param('channel', type=int, choices=[1, 2, 3])
    @ocs_agent.param('volts', type=float, check=lambda x: 0 <= x <= 30)
    def set_voltage(self, session, params=None):
        """set_voltage(channel, volts)

        **Task** - Set the voltage of the power supply.

        Parameters:
            channel (int): Channel number (1, 2, or 3).
            volts (float): Voltage to set. Must be between 0 and 30.

        """

        with self.lock.acquire_timeout(1) as acquired:
            if acquired:
                self.psu.set_volt(params['channel'], params['volts'])
            else:
                return False, "Could not acquire lock"

        return True, 'Set channel {} voltage to {}'.format(params['channel'], params['volts'])

    @ocs_agent.param('channel', type=int, choices=[1, 2, 3])
    @ocs_agent.param('current', type=float)
    def set_current(self, session, params=None):
        """set_current(channel, current)

        **Task** - Set the current of the power supply.

        Parameters:
            channel (int): Channel number (1, 2, or 3).
            current (float): Current to set.

        """
        with self.lock.acquire_timeout(1) as acquired:
            if acquired:
                self.psu.set_curr(params['channel'], params['current'])
            else:
                return False, "Could not acquire lock"

        return True, 'Set channel {} current to {}'.format(params['channel'], params['current'])

    @ocs_agent.param('channel', type=int, choices=[1, 2, 3])
    @ocs_agent.param('state', type=bool)
    def set_output(self, session, params=None):
        """set_output(channel, state)

        **Task** - Turn a channel on or off.

        Parameters:
            channel (int): Channel number (1, 2, or 3).
            state (bool): True for on, False for off.

        """
        with self.lock.acquire_timeout(1) as acquired:
            if acquired:
                self.psu.set_output(params['channel'], params['state'])
            else:
                return False, "Could not acquire lock"

        return True, 'Initialized PSU.'


def make_parser(parser=None):
    """Build the argument parser for the Agent. Allows sphinx to automatically
    build documentation based on this function.

    """
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent.
    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--ip-address')
    pgroup.add_argument('--gpib-slot')

    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args(agent_class='ScpiPsuAgent',
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)

    p = ScpiPsuAgent(agent, args.ip_address, int(args.gpib_slot))

    agent.register_task('init', p.init)
    agent.register_task('set_voltage', p.set_voltage)
    agent.register_task('set_current', p.set_current)
    agent.register_task('set_output', p.set_output)

    agent.register_process('monitor_output', p.monitor_output, p.stop_monitoring)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import argparse
import time

from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock
from twisted.internet import reactor

import socs.agents.hwp_rotation.drivers.pid_controller as pd
from socs.common.pmx import PMX, Command


class RotationAgent:
    """Agent to control the rotation speed of the CHWP

    Args:
        kikusui_ip (str): IP address for the Kikusui power supply
        kikusui_port (str): Port for the Kikusui power supply
        pid_ip (str): IP address for the PID controller
        pid_port (str): Port for the PID controller
        pid_verbosity (str): Verbosity of PID controller output

    """

    def __init__(self, agent, kikusui_ip, kikusui_port, pid_ip, pid_port, pid_verbosity):
        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()
        self._initialized = False
        self.take_data = False
        self.kikusui_ip = kikusui_ip
        self.kikusui_port = int(kikusui_port)
        self.pid_ip = pid_ip
        self.pid_port = pid_port
        self._pid_verbosity = pid_verbosity > 0
        self.cmd = None  # Command object for PSU commanding
        self.pid = None  # PID object for pid controller commanding

        agg_params = {'frame_length': 60}
        self.agent.register_feed(
            'hwprotation', record=True, agg_params=agg_params)

    @ocs_agent.param('auto_acquire', default=False, type=bool)
    @ocs_agent.param('force', default=False, type=bool)
    def init_connection(self, session, params):
        """init_connection(auto_acquire=False, force=False)

        **Task** - Initialize connection to Kikusui Power Supply and PID
        Controller.

        Parameters:
            auto_acquire (bool, optional): Default is False. Starts data
                acquisition after initialization if True.
            force (bool, optional): Force initialization, even if already
                initialized. Defaults to False.

        """
        if self._initialized and not params['force']:
            self.log.info("Connection already initialized. Returning...")
            return True, "Connection already initialized"

        with self.lock.acquire_timeout(0, job='init_connection') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not run init_connection because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'

            try:
                pmx = PMX(tcp_ip=self.kikusui_ip,
                          tcp_port=self.kikusui_port, timeout=0.5)
                self.cmd = Command(pmx)
                self.log.info('Connected to Kikusui power supply')
            except ConnectionRefusedError:
                self.log.error('Could not establish connection to Kikusui power supply')
                reactor.callFromThread(reactor.stop)
                return False, 'Unable to connect to Kikusui PSU'

            try:
                self.pid = pd.PID(pid_ip=self.pid_ip, pid_port=self.pid_port,
                                  verb=self._pid_verbosity)
                self.log.info('Connected to PID controller')
            except BrokenPipeError:
                self.log.error('Could not establish connection to PID controller')
                reactor.callFromThread(reactor.stop)
                return False, 'Unable to connect to PID controller'

        self._initialized = True

        # Start 'iv_acq' Process if requested
        if params['auto_acquire']:
            self.agent.start('iv_acq')

        return True, 'Connection to PSU and PID controller established'

    def tune_stop(self, session, params):
        """tune_stop()

        **Task** - Reverse the drive direction of the PID controller and
        optimize the PID parameters for deceleration.

        """
        with self.lock.acquire_timeout(3, job='tune_stop') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not tune stop because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'

            self.pid.tune_stop()

        return True, 'Reversing Direction'

    def tune_freq(self, session, params):
        """tune_freq()

        **Task** - Tune the PID controller setpoint to the rotation frequency
        and optimize the PID parameters for rotation.

        """
        with self.lock.acquire_timeout(3, job='tune_freq') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not tune freq because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'

            self.pid.tune_freq()

        return True, 'Tuning to setpoint'

    @ocs_agent.param('freq', default=0., check=lambda x: 0. <= x <= 3.0)
    def declare_freq(self, session, params):
        """declare_freq(freq=0)

        **Task** - Store the entered frequency as the PID setpoint when
        ``tune_freq()`` is next called.

        Parameters:
            freq (float): Desired HWP rotation frequency

        """
        with self.lock.acquire_timeout(3, job='declare_freq') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not declare freq because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'

            self.pid.declare_freq(params['freq'])

        return True, 'Setpoint at {} Hz'.format(params['freq'])

    @ocs_agent.param('p', default=0.2, type=float, check=lambda x: 0. < x <= 8.)
    @ocs_agent.param('i', default=63, type=int, check=lambda x: 0 <= x <= 200)
    @ocs_agent.param('d', default=0., type=float, check=lambda x: 0. <= x < 10.)
    def set_pid(self, session, params):
        """set_pid(p=0.2, i=63, d=0.)

        **Task** - Set the PID parameters. Note these changes are for the
        current session only and will change whenever the agent container is
        reloaded.

        Parameters:
            p (float): Proportional PID value
            i (int): Integral PID value
            d (float): Derivative PID value

        """
        with self.lock.acquire_timeout(3, job='set_pid') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not set pid because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'

            self.pid.set_pid(
                [params['p'], params['i'], params['d']])

        return True, f"Set PID params to p: {params['p']}, i: {params['i']}, d: {params['d']}"

    def get_freq(self, session, params):
        """get_freq()

        **Task** - Return the current HWP frequency as seen by the PID
        controller.

        """
        with self.lock.acquire_timeout(3, job='get_freq') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not get freq because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'

            freq = self.pid.get_freq()

        return True, 'Current frequency = {}'.format(freq)

    def get_direction(self, session, params):
        """get_direction()

        **Task** - Return the current HWP tune direction as seen by the PID
        controller.

        """
        with self.lock.acquire_timeout(3, job='get_direction') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not get freq because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'

            direction = self.pid.get_direction()

        return True, 'Current Direction = {}'.format(['Forward', 'Reverse'][direction])

    @ocs_agent.param('direction', type=str, default='0', choices=['0', '1'])
    def set_direction(self, session, params):
        """set_direction(direction='0')

        **Task** - Set the HWP rotation direction.

        Parameters:
            direction (str): '0' for forward and '1' for reverse.

        """
        with self.lock.acquire_timeout(3, job='set_direction') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not set direction because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'

            self.pid.set_direction(params['direction'])

        return True, 'Set direction'

    @ocs_agent.param('slope', default=1., type=float, check=lambda x: -10. < x < 10.)
    @ocs_agent.param('offset', default=0.1, type=float, check=lambda x: -10. < x < 10.)
    def set_scale(self, session, params):
        """set_scale(slope=1, offset=0.1)

        **Task** - Set the PID's internal conversion from input voltage to
        rotation frequency.

        Parameters:
            slope (float): Slope of the "rotation frequency vs input voltage"
                relationship
            offset (float): y-intercept of the "rotation frequency vs input
                voltage" relationship

        """
        with self.lock.acquire_timeout(3, job='set_scale') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not set scale because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'

            self.pid.set_scale(params['slope'], params['offset'])

        return True, 'Set scale'

    def set_on(self, session, params):
        """set_on()

        **Task** - Turn on the Kikusui drive voltage.

        """
        with self.lock.acquire_timeout(3, job='set_on') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not set on because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'

            time.sleep(1)
            self.cmd.user_input('on')

        return True, 'Set Kikusui on'

    def set_off(self, session, params):
        """set_off()

        **Task** - Turn off the Kikusui drive voltage.

        """
        with self.lock.acquire_timeout(3, job='set_off') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not set off because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'

            time.sleep(1)
            self.cmd.user_input('off')

        return True, 'Set Kikusui off'

    @ocs_agent.param('volt', default=0, type=float, check=lambda x: 0 <= x <= 35)
    def set_v(self, session, params):
        """set_v(volt=0)

        **Task** - Set the Kikusui drive voltage.

        Parameters:
            volt (float): Kikusui set voltage

        """
        with self.lock.acquire_timeout(3, job='set_v') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not set v because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'

            time.sleep(1)
            self.cmd.user_input('V {}'.format(params['volt']))

        return True, 'Set Kikusui voltage to {} V'.format(params['volt'])

    @ocs_agent.param('volt', default=32., type=float, check=lambda x: 0. <= x <= 35.)
    def set_v_lim(self, session, params):
        """set_v_lim(volt=32)

        **Task** - Set the Kikusui drive voltage limit.

        Parameters:
            volt (float): Kikusui limit voltage

        """
        with self.lock.acquire_timeout(3, job='set_v_lim') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not set v lim because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'

            time.sleep(1)
            print(params['volt'])
            self.cmd.user_input('VL {}'.format(params['volt']))

        return True, 'Set Kikusui voltage limit to {} V'.format(params['volt'])

    def use_ext(self, session, params):
        """use_ext()

        **Task** - Set the Kikusui to use an external voltage control. Doing so
        enables PID control.

        """
        with self.lock.acquire_timeout(3, job='use_ext') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not use external voltage because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'

            time.sleep(1)
            self.cmd.user_input('U')

        return True, 'Set Kikusui voltage to PID control'

    def ign_ext(self, session, params):
        """ign_ext()

        **Task** - Set the Kiksui to ignore external voltage control. Doing so
        disables the PID and switches to direct control.

        """
        with self.lock.acquire_timeout(3, job='ign_ext') as acquired:
            if not acquired:
                self.log.warn(
                    'Could not ignore external voltage because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'

            time.sleep(1)
            self.cmd.user_input('I')

        return True, 'Set Kikusui voltage to direct control'

    @ocs_agent.param('test_mode', default=False, type=bool)
    def iv_acq(self, session, params):
        """iv_acq(test_mode=False)

        **Process** - Start Kikusui data acquisition.

        Parameters:
            test_mode (bool, optional): Run the Process loop only once.
                This is meant only for testing. Default is False.

        Notes:
            The most recent data collected is stored in the session data in the
            structure::

                >>> response.session['data']
                {'kikusui_volt': 0,
                 'kikusui_curr': 0,
                 'last_updated': 1649085992.719602}

        """
        with self.lock.acquire_timeout(timeout=0, job='iv_acq') as acquired:
            if not acquired:
                self.log.warn('Could not start iv acq because {} is already running'
                              .format(self.lock.job))
                return False, 'Could not acquire lock'

            session.set_status('running')
            last_release = time.time()
            self.take_data = True

            while self.take_data:
                # Relinquish sampling lock occasionally.
                if time.time() - last_release > 1.:
                    last_release = time.time()
                    if not self.lock.release_and_acquire(timeout=10):
                        self.log.warn(f"Failed to re-acquire sampling lock, "
                                      f"currently held by {self.lock.job}.")
                        continue

                data = {'timestamp': time.time(),
                        'block_name': 'HWPKikusui_IV', 'data': {}}

                v_msg, v_val = self.cmd.user_input('V?')
                i_msg, i_val = self.cmd.user_input('C?')

                data['data']['kikusui_volt'] = v_val
                data['data']['kikusui_curr'] = i_val

                self.agent.publish_to_feed('hwprotation', data)

                session.data = {'kikusui_volt': v_val,
                                'kikusui_curr': i_val,
                                'last_updated': time.time()}

                time.sleep(1)

                if params['test_mode']:
                    break

        self.agent.feeds['hwprotation'].flush_buffer()
        return True, 'Acqusition exited cleanly'

    def _stop_iv_acq(self, session, params):
        """
        Stop iv_acq process.
        """
        if self.take_data:
            self.take_data = False
            return True, 'requested to stop taking data'

        return False, 'acq is not currently running'


def make_parser(parser=None):
    """
    Build the argument parser for the Agent. Allows sphinx to automatically build documentation
    baised on this function
    """
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent
    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--kikusui-ip')
    pgroup.add_argument('--kikusui-port')
    pgroup.add_argument('--pid-ip')
    pgroup.add_argument('--pid-port')
    pgroup.add_argument('--verbose', '-v', action='count', default=0,
                        help='PID Controller verbosity level.')
    pgroup.add_argument('--mode', type=str, default='iv_acq',
                        choices=['idle', 'init', 'iv_acq'],
                        help="Starting operation for the Agent.")
    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args(agent_class='RotationAgent',
                                  parser=parser,
                                  args=args)

    init_params = False
    if args.mode == 'init':
        init_params = {'auto_acquire': False}
    elif args.mode == 'iv_acq':
        init_params = {'auto_acquire': True}

    agent, runner = ocs_agent.init_site_agent(args)
    rotation_agent = RotationAgent(agent, kikusui_ip=args.kikusui_ip,
                                   kikusui_port=args.kikusui_port,
                                   pid_ip=args.pid_ip,
                                   pid_port=args.pid_port,
                                   pid_verbosity=args.verbose)
    agent.register_process('iv_acq', rotation_agent.iv_acq,
                           rotation_agent._stop_iv_acq)
    agent.register_task('init_connection', rotation_agent.init_connection,
                        startup=init_params)
    agent.register_task('tune_stop', rotation_agent.tune_stop)
    agent.register_task('tune_freq', rotation_agent.tune_freq)
    agent.register_task('declare_freq', rotation_agent.declare_freq)
    agent.register_task('set_pid', rotation_agent.set_pid)
    agent.register_task('get_freq', rotation_agent.get_freq)
    agent.register_task('get_direction', rotation_agent.get_direction)
    agent.register_task('set_direction', rotation_agent.set_direction)
    agent.register_task('set_scale', rotation_agent.set_scale)
    agent.register_task('set_on', rotation_agent.set_on)
    agent.register_task('set_off', rotation_agent.set_off)
    agent.register_task('set_v', rotation_agent.set_v)
    agent.register_task('set_v_lim', rotation_agent.set_v_lim)
    agent.register_task('use_ext', rotation_agent.use_ext)
    agent.register_task('ign_ext', rotation_agent.ign_ext)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



# Author: Michael Randall
# Email: mrandall@ucsd.edu

# This Driver works by creating a TCP connection to a Moxa Ethernet to Serial Converter.
# It uses this Converter to send and receive serial messages with the Pfeiffer Vacuum controller.
# The Driver employs the serial package to creat the TCP connection
# It also uses a slightly modified version of a Pfeiffer Vacuum Protocol package found on GitHub

import serial
from pfeiffer_vacuum_protocol.pfeiffer_vacuum_protocol import \
    _read_gauge_response as read_gauge_response
from pfeiffer_vacuum_protocol.pfeiffer_vacuum_protocol import \
    _send_control_command as send_control_command
from pfeiffer_vacuum_protocol.pfeiffer_vacuum_protocol import \
    _send_data_request as send_data_request

# Data type 0 from TC400 Manual Section 8.3 - Applied data types
PFEIFFER_BOOL = {'111111': True,
                 '000000': False}


class PfeifferTC400:
    """Initiates a TCP connection with the Moxa serial to ethernet converter to send serial communications.

    Parameters
    ----------
        moxa_ip_address: str
            The IP address of the moxa box
        moxa_port: int
            The port number of the Moxa box that the turbo is connected to.
            (e.g. 4001 for the first port)
        turbo_address: int
            The serial address of the turbo controller (e.g. 94)
            Check the turbo for the address.

    Attributes
    ----------
        ser: serial.Serial Object
            The TCP connection with the Moxa used to send and receive communication.
        turbo_address: int
            The serial Address of the Turbo Controller.
    """

    def __init__(self, moxa_ip_address, moxa_port, turbo_address):
        self.ser = serial.serial_for_url('socket://{}:{}'.format(moxa_ip_address, moxa_port),
                                         baudrate=9600,
                                         bytesize=serial.EIGHTBITS,
                                         parity=serial.PARITY_NONE,
                                         stopbits=serial.STOPBITS_ONE,
                                         timeout=3)

        self.turbo_address = turbo_address

    def get_turbo_motor_temperature(self):
        """Gets the temperatures of the turbo rotor from the turbo controller.

        Returns
        -------
        int
            The rotor temperature of the turbo in Celsius.
        """

        send_data_request(self.ser, self.turbo_address, 346)
        addr, rw, param_num, motor_temp = read_gauge_response(self.ser)

        return int(motor_temp)

    def get_turbo_actual_rotation_speed(self):
        """Gets the current rotation speed of the turbo from the turbo controller.

        Returns
        -------
        int
            The current rotation speed of the turbo in Hz.
        """

        send_data_request(self.ser, self.turbo_address, 309)

        addr, rw, param_num, actual_rotation_speed = read_gauge_response(self.ser)

        return int(actual_rotation_speed)

    def get_turbo_set_rotation_speed(self):
        """Gets the the rotation speed that the turbo is set to from the turbo controller.
        This is the speed in Hz that the turbo motor will spin up to if turned on.

        Returns
        -------
        int
            The rotation speed that the turbo is set to in Hz
        """

        send_data_request(self.ser, self.turbo_address, 308)

        addr, rw, param_num, set_rotation_speed = read_gauge_response(self.ser)

        return int(set_rotation_speed)

    def get_turbo_error_code(self):
        """Gets the current error code of the turbo from the turbo controller.

        Returns
        -------
        str
            The current error code of the turbo.
        """
        send_data_request(self.ser, self.turbo_address, 303)

        addr, rw, param_num, error_code = read_gauge_response(self.ser)

        return error_code

    def unready_turbo(self):
        """Unreadies the turbo. Does not cause the turbo to spin up.
        Returns
        -------
        bool
            True for successful, False for failure.
        """

        send_control_command(self.ser, self.turbo_address, 10, "000000")

        addr, rw, param_num, turbo_response = read_gauge_response(self.ser)

        if turbo_response not in PFEIFFER_BOOL:
            raise ValueError(f"Unrecognized response from turbo: {turbo_response}")
        else:
            return turbo_response == "111111"

    def ready_turbo(self):
        """Readies the turbo for spinning. Does not cause the turbo to spin up.

        Returns
        -------
        bool
            True for successful, False for failure.
        """

        send_control_command(self.ser, self.turbo_address, 10, "111111")

        addr, rw, param_num, turbo_response = read_gauge_response(self.ser)

        if turbo_response not in PFEIFFER_BOOL:
            raise ValueError(f"Unrecognized response from turbo: {turbo_response}")
        else:
            return turbo_response == "111111"

    def turn_turbo_motor_on(self):
        """Turns the turbo motor on. The turbo must be readied before the motor will turn on.
        This is what causes the turbo to actually spin up.

        Returns
        -------
        bool
            True for successful, False for failure.
        """

        send_control_command(self.ser, self.turbo_address, 23, "111111")

        addr, rw, param_num, turbo_response = read_gauge_response(self.ser)

        if turbo_response not in PFEIFFER_BOOL:
            raise ValueError(f"Unrecognized response from turbo: {turbo_response}")
        else:
            return turbo_response == "111111"

    def turn_turbo_motor_off(self):
        """Turns the turbo motor off.

        Returns
        -------
        bool
            True for successful, False for failure.
        """

        send_control_command(self.ser, self.turbo_address, 23, "000000")

        addr, rw, param_num, turbo_response = read_gauge_response(self.ser)

        if turbo_response not in PFEIFFER_BOOL:
            raise ValueError(f"Unrecognized response from turbo: {turbo_response}")
        else:
            return turbo_response == "111111"

    def acknowledge_turbo_errors(self):
        """Acknowledges the turbo errors. This is analagous to clearing the errors.
        If the errors were fixed, the turbo will be able to be turned back on.

        Returns
        -------
        bool
            True for successful, False for failure.
        """

        send_control_command(self.ser, self.turbo_address, 9, "111111")

        addr, rw, param_num, turbo_response = read_gauge_response(self.ser)

        if turbo_response not in PFEIFFER_BOOL:
            raise ValueError(f"Unrecognized response from turbo: {turbo_response}")
        else:
            return turbo_response == "111111"



import argparse
import os
import time

import txaio
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock

from socs.Lakeshore import Lakeshore425 as ls

txaio.use_twisted()


class LS425Agent:
    """Agent for interfacing with a single Lakeshore 425 device.

    Args:
        agent (ocs.ocs_agent.OCSAgent): Instantiated OCSAgent class for this Agent
        port (int): Path to USB device in `/dev/`
        f_sample (float): Default sampling rate for the acq Process

    """

    def __init__(self, agent, port, f_sample=1.):

        self.agent: ocs_agent.OCSAgent = agent
        self.log = agent.log
        self.lock = TimeoutLock()

        self.port = port
        self.dev = None

        self.f_sample = f_sample

        self.initialized = False
        self.take_data = False

        # Registers Temperature and Voltage feeds
        agg_params = {'frame_length': 60}
        self.agent.register_feed('mag_field',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=1)

    # Task functions.
    @ocs_agent.param('auto_acquire', default=False, type=bool)
    def init_lakeshore(self, session, params):
        """init_lakeshore(auto_acquire=False)

        **Task** - Perform first time setup of the Lakeshore 425 Module.

        Parameters:
            auto_acquire (bool, optional): Default is False. Starts data
                acquisition after initialization if True.

        """
        if params is None:
            params = {}

        auto_acquire = params['auto_acquire']

        if self.initialized:
            return True, "Already Initialized Module"

        with self.lock.acquire_timeout(0, job='init') as acquired:
            if not acquired:
                self.log.warn("Could not start init because "
                              "{} is already running".format(self.lock.job))
                return False, "Could not acquire lock."

            session.set_status('starting')

            self.dev = ls.LakeShore425(self.port)
            self.log.info(self.dev.get_id())
            print("Initialized Lakeshore module: {!s}".format(self.dev))

        self.initialized = True
        # Start data acquisition if requested
        if auto_acquire:
            self.agent.start('acq', params={'sampling_frequency': self.f_sample})

        return True, 'Lakeshore module initialized.'

    @ocs_agent.param('sampling_frequency', default=None, type=float)
    def acq(self, session, params):
        """acq(sampling_frequency=None)

        **Process** - Acquire data from the Lakeshore 425.

        Parameters:
            sampling_frequency (float, optional):
                Sampling frequency for data collection. Defaults to the value
                passed to `--sampling_frequency` on Agent startup

        Notes:
            The most recent data collected is stored in session data in the
            structure::

                >>> response.session['data']
                {"fields":
                    {"mag_field": {"Bfield": 270.644},
                     "timestamp": 1601924466.6130798}
                }

        """
        if params is None:
            params = {}

        f_sample = params['sampling_frequency']
        # If f_sample is None, use value passed to Agent init
        if f_sample is None:
            f_sample = self.f_sample

        sleep_time = 1 / f_sample - 0.01

        with self.lock.acquire_timeout(0, job='acq') as acquired:
            if not acquired:
                self.log.warn("Could not start acq because {} is already running"
                              .format(self.lock.job))
                return False, "Could not acquire lock."

            session.set_status('running')

            self.take_data = True

            session.data = {"fields": {}}

            last_release = time.time()
            while self.take_data:
                if time.time() - last_release > 1.:
                    last_release = time.time()
                    if not self.lock.release_and_acquire(timeout=10):
                        self.log.warn(f"Failed to re-acquire lock, currently held by {self.lock.job}.")
                        continue

                Bfield = self.dev.get_field()
                current_time = time.time()
                data = {
                    'timestamp': current_time,
                    'block_name': 'mag_field',
                    'data': {'Bfield': Bfield}
                }

                self.agent.publish_to_feed('mag_field', data)
                session.data.update({'timestamp': current_time})
                self.agent.feeds['mag_field'].flush_buffer()

                time.sleep(sleep_time)

        return True, 'Acquisition exited cleanly.'

    def _stop_acq(self, session, params):
        """
        Stops acq process.
        """
        if self.take_data:
            self.take_data = False
            return True, 'requested to stop taking data.'
        else:
            return False, 'acq is not currently running'

    @ocs_agent.param('_')
    def operational_status(self, session, params):
        """operational_status()

        **Task** - Check operational status.

        """
        with self.lock.acquire_timeout(3, job='operational_status') as acquired:
            if not acquired:
                self.log.warn('Could not start operational_status because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'
            op_status = self.dev.get_op_status()
            self.log.info(op_status)
            return True, 'operational status: ' + op_status

    @ocs_agent.param('_')
    def zero_calibration(self, session, params):
        """zero_calibration()

        **Task** - Calibrate the zero point.

        """
        with self.lock.acquire_timeout(3, job='zero_calibration') as acquired:
            if not acquired:
                self.log.warn('Could not start zero_calibration because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'
            self.dev.set_zero()
            return True, 'Zero calibration is done'

    @ocs_agent.param('command', type=str)
    def any_command(self, session, params):
        """any_command(command)

        **Process** - Send serial command to Lakeshore 425

        Parameters:
            command (str): any serial command

        Examples:
            Example for calling in a client::

                >>> client.any_command(command='*IDN?')

        Notes:
            An example of the session data::

                >>> response.session['data']
                {'response': 'LSA1234'}

        """
        command = params['command']
        with self.lock.acquire_timeout(3, job='any_command') as acquired:
            if not acquired:
                self.log.warn('Could not any_command because {} is already running'.format(self.lock.job))
                return False, 'Could not acquire lock'
            print('Input command: ' + command)
            if '?' in command:
                out = self.dev.query(command)
                session.data = {'response': out}
                return True, 'any_command is finished cleanly. Results: {}'.format(out)
            else:
                self.dev.command(command)
                session.data = {'response': None}
                return True, 'any_command is finished cleanly'


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--port', type=str,
                        help="Path to USB node for the lakeshore")
    pgroup.add_argument('--mode', type=str, choices=['init', 'acq'],
                        help="Starting action for the agent.")
    pgroup.add_argument('--sampling-frequency', type=float,
                        help="Sampling frequency for data acquisition")
    return parser


def main(args=None):
    # Start logging
    txaio.start_logging(level=os.environ.get("LOGLEVEL", "info"))

    parser = make_parser()
    args = site_config.parse_args(agent_class='Lakeshore425Agent',
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)

    # Automatically acquire data if requested (default)
    init_params = False
    if args.mode == 'init':
        init_params = {'auto_acquire': False}
    elif args.mode == 'acq':
        init_params = {'auto_acquire': True}

    kwargs = {'port': args.port}

    if args.sampling_frequency is not None:
        kwargs['f_sample'] = float(args.sampling_frequency)
    gauss = LS425Agent(agent, **kwargs)

    agent.register_task('init_lakeshore', gauss.init_lakeshore, startup=init_params)
    agent.register_task('operational_status', gauss.operational_status)
    agent.register_task('zero_calibration', gauss.zero_calibration)
    agent.register_task('any_command', gauss.any_command)
    agent.register_process('acq', gauss.acq, gauss._stop_acq)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



#
# PySNMP MIB module IBOOTPDU-MIB (http://snmplabs.com/pysmi)
# ASN.1 source file://./iboot-pdu.mib
# Produced by pysmi-0.3.4 at Fri May 27 15:35:01 2022
# On host HAWKING platform Linux version 5.10.16.3-microsoft-standard-WSL2 by user davidvng
# Using Python version 3.8.8 (default, Apr 13 2021, 19:58:26)
#
ObjectIdentifier, OctetString, Integer = mibBuilder.importSymbols("ASN1", "ObjectIdentifier", "OctetString", "Integer")
NamedValues, = mibBuilder.importSymbols("ASN1-ENUMERATION", "NamedValues")
ConstraintsUnion, ValueRangeConstraint, ValueSizeConstraint, ConstraintsIntersection, SingleValueConstraint = mibBuilder.importSymbols("ASN1-REFINEMENT", "ConstraintsUnion", "ValueRangeConstraint", "ValueSizeConstraint", "ConstraintsIntersection", "SingleValueConstraint")
ModuleCompliance, NotificationGroup = mibBuilder.importSymbols("SNMPv2-CONF", "ModuleCompliance", "NotificationGroup")
TimeTicks, enterprises, Bits, IpAddress, Unsigned32, MibScalar, MibTable, MibTableRow, MibTableColumn, ObjectIdentity, Integer32, Counter64, iso, Counter32, NotificationType, ModuleIdentity, MibIdentifier, Gauge32 = mibBuilder.importSymbols("SNMPv2-SMI", "TimeTicks", "enterprises", "Bits", "IpAddress", "Unsigned32", "MibScalar", "MibTable", "MibTableRow", "MibTableColumn", "ObjectIdentity", "Integer32", "Counter64", "iso", "Counter32", "NotificationType", "ModuleIdentity", "MibIdentifier", "Gauge32")
DisplayString, TextualConvention, MacAddress = mibBuilder.importSymbols("SNMPv2-TC", "DisplayString", "TextualConvention", "MacAddress")
iBoot_PDU_Agent = ModuleIdentity((1, 3, 6, 1, 4, 1, 1418, 6)).setLabel("iBoot-PDU-Agent")
iBoot_PDU_Agent.setRevisions(('2017-10-25 13:23',))
if mibBuilder.loadTexts: iBoot_PDU_Agent.setLastUpdated('201710251323Z')
if mibBuilder.loadTexts: iBoot_PDU_Agent.setOrganization('')
class TC1(TextualConvention, Integer32):
    status = 'current'

dataprobe = MibIdentifier((1, 3, 6, 1, 4, 1, 1418))
device = MibIdentifier((1, 3, 6, 1, 4, 1, 1418, 6, 1))
console = MibIdentifier((1, 3, 6, 1, 4, 1, 1418, 6, 2))
network = MibIdentifier((1, 3, 6, 1, 4, 1, 1418, 6, 3))
snmp = MibIdentifier((1, 3, 6, 1, 4, 1, 1418, 6, 4))
sensors = MibIdentifier((1, 3, 6, 1, 4, 1, 1418, 6, 6))
notifications = MibIdentifier((1, 3, 6, 1, 4, 1, 1418, 6, 7))
firmwareVersion = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 1, 1), DisplayString()).setMaxAccess("readonly")
if mibBuilder.loadTexts: firmwareVersion.setStatus('current')
deviceName = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 1, 2), DisplayString()).setMaxAccess("readwrite")
if mibBuilder.loadTexts: deviceName.setStatus('current')
deviceFamily = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 1, 3), DisplayString()).setMaxAccess("readonly")
if mibBuilder.loadTexts: deviceFamily.setStatus('current')
deviceModelName = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 1, 4), DisplayString()).setMaxAccess("readonly")
if mibBuilder.loadTexts: deviceModelName.setStatus('current')
deviceConnector = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 1, 5), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(0, 1))).clone(namedValues=NamedValues(("nema", 0), ("iec", 1)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: deviceConnector.setStatus('current')
deviceNumberOfOutlets = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 1, 6), Integer32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: deviceNumberOfOutlets.setStatus('current')
deviceNumberOfLineCords = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 1, 7), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(0, 1))).clone(namedValues=NamedValues(("one", 0), ("two", 1)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: deviceNumberOfLineCords.setStatus('current')
deviceMaxCurrent = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 1, 8), Integer32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: deviceMaxCurrent.setStatus('current')
deviceTemperatureUnit = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 1, 9), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(0, 1))).clone(namedValues=NamedValues(("fahrenheit", 0), ("celsius", 1)))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: deviceTemperatureUnit.setStatus('current')
deviceTimeZone = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 1, 10), DisplayString()).setMaxAccess("readonly")
if mibBuilder.loadTexts: deviceTimeZone.setStatus('current')
deviceCalibrated = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 1, 11), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(0, 1))).clone(namedValues=NamedValues(("false", 0), ("true", 1)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: deviceCalibrated.setStatus('current')
modemCountryCode = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 1, 12), Integer32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: modemCountryCode.setStatus('current')
outletDelayTime = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 1, 13), Integer32()).setMaxAccess("readwrite")
if mibBuilder.loadTexts: outletDelayTime.setStatus('current')
cloudServiceEnabled = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 1, 14), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(0, 1))).clone(namedValues=NamedValues(("false", 0), ("true", 1)))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: cloudServiceEnabled.setStatus('current')
cloudServerAddress = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 1, 15), DisplayString().subtype(subtypeSpec=ValueSizeConstraint(128, 128)).setFixedLength(128)).setMaxAccess("readwrite")
if mibBuilder.loadTexts: cloudServerAddress.setStatus('current')
cloudActivationCode = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 1, 16), DisplayString()).setMaxAccess("readonly")
if mibBuilder.loadTexts: cloudActivationCode.setStatus('current')
cloudUUID = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 1, 17), DisplayString()).setMaxAccess("readonly")
if mibBuilder.loadTexts: cloudUUID.setStatus('current')
setFactoryDefaults = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 1, 18), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(0, 1))).clone(namedValues=NamedValues(("false", 0), ("true", 1)))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: setFactoryDefaults.setStatus('current')
rebootSystem = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 1, 19), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(0, 1))).clone(namedValues=NamedValues(("false", 0), ("true", 1)))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: rebootSystem.setStatus('current')
rebootRequired = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 1, 20), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(0, 1))).clone(namedValues=NamedValues(("false", 0), ("true", 1)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: rebootRequired.setStatus('current')
serialPortEnabled = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 2, 1), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(0, 1))).clone(namedValues=NamedValues(("false", 0), ("true", 1)))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: serialPortEnabled.setStatus('current')
consoleTimeout = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 2, 2), Integer32().subtype(subtypeSpec=ValueRangeConstraint(0, 60))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: consoleTimeout.setStatus('current')
telnetEnabled = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 2, 3), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(0, 1))).clone(namedValues=NamedValues(("false", 0), ("true", 1)))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: telnetEnabled.setStatus('current')
telnetPort = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 2, 4), Integer32().subtype(subtypeSpec=ValueRangeConstraint(1, 65535))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: telnetPort.setStatus('current')
sshEnabled = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 2, 5), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(0, 1))).clone(namedValues=NamedValues(("false", 0), ("true", 1)))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: sshEnabled.setStatus('current')
sshPort = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 2, 6), Integer32().subtype(subtypeSpec=ValueRangeConstraint(1, 65535))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: sshPort.setStatus('current')
macAddress = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 3, 1), MacAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: macAddress.setStatus('current')
ipMode = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 3, 2), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(0, 1))).clone(namedValues=NamedValues(("static", 0), ("dhcp", 1)))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: ipMode.setStatus('current')
ipAddress = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 3, 3), IpAddress()).setMaxAccess("readwrite")
if mibBuilder.loadTexts: ipAddress.setStatus('current')
subnetMask = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 3, 4), IpAddress()).setMaxAccess("readwrite")
if mibBuilder.loadTexts: subnetMask.setStatus('current')
gateway = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 3, 5), IpAddress()).setMaxAccess("readwrite")
if mibBuilder.loadTexts: gateway.setStatus('current')
dnsServer1 = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 3, 6), IpAddress()).setMaxAccess("readwrite")
if mibBuilder.loadTexts: dnsServer1.setStatus('current')
dnsServer2 = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 3, 7), IpAddress()).setMaxAccess("readwrite")
if mibBuilder.loadTexts: dnsServer2.setStatus('current')
sslEnabled = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 3, 8), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(0, 1))).clone(namedValues=NamedValues(("false", 0), ("true", 1)))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: sslEnabled.setStatus('current')
sslPort = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 3, 9), Integer32().subtype(subtypeSpec=ValueRangeConstraint(1, 65535))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: sslPort.setStatus('current')
webEnabled = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 3, 10), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(0, 1))).clone(namedValues=NamedValues(("false", 0), ("true", 1)))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: webEnabled.setStatus('current')
webPort = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 3, 11), Integer32().subtype(subtypeSpec=ValueRangeConstraint(0, 65535))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: webPort.setStatus('current')
snmpEnabled = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 4, 1), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(0, 1))).clone(namedValues=NamedValues(("false", 0), ("true", 1)))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: snmpEnabled.setStatus('current')
snmpReadCommunity = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 4, 2), DisplayString()).setMaxAccess("readwrite")
if mibBuilder.loadTexts: snmpReadCommunity.setStatus('current')
snmpWriteCommunity = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 4, 3), DisplayString()).setMaxAccess("readwrite")
if mibBuilder.loadTexts: snmpWriteCommunity.setStatus('current')
snmpManagerTable = MibTable((1, 3, 6, 1, 4, 1, 1418, 6, 4, 4), )
if mibBuilder.loadTexts: snmpManagerTable.setStatus('current')
snmpManagerEntry = MibTableRow((1, 3, 6, 1, 4, 1, 1418, 6, 4, 4, 1), ).setIndexNames((0, "IBOOTPDU-MIB", "snmpManagerIndex"))
if mibBuilder.loadTexts: snmpManagerEntry.setStatus('current')
snmpManagerIndex = MibTableColumn((1, 3, 6, 1, 4, 1, 1418, 6, 4, 4, 1, 1), Integer32().subtype(subtypeSpec=ValueRangeConstraint(0, 255))).setMaxAccess("readonly")
if mibBuilder.loadTexts: snmpManagerIndex.setStatus('current')
snmpManagerAddress = MibTableColumn((1, 3, 6, 1, 4, 1, 1418, 6, 4, 4, 1, 2), DisplayString().subtype(subtypeSpec=ValueSizeConstraint(16, 16)).setFixedLength(16)).setMaxAccess("readwrite")
if mibBuilder.loadTexts: snmpManagerAddress.setStatus('current')
snmpManagerName = MibTableColumn((1, 3, 6, 1, 4, 1, 1418, 6, 4, 4, 1, 3), DisplayString()).setMaxAccess("readwrite")
if mibBuilder.loadTexts: snmpManagerName.setStatus('current')
snmpTrapCommunity = MibTableColumn((1, 3, 6, 1, 4, 1, 1418, 6, 4, 4, 1, 4), DisplayString()).setMaxAccess("readwrite")
if mibBuilder.loadTexts: snmpTrapCommunity.setStatus('current')
outletTable = MibTable((1, 3, 6, 1, 4, 1, 1418, 6, 5), )
if mibBuilder.loadTexts: outletTable.setStatus('current')
outletEntry = MibTableRow((1, 3, 6, 1, 4, 1, 1418, 6, 5, 1), ).setIndexNames((0, "IBOOTPDU-MIB", "outletIndex"))
if mibBuilder.loadTexts: outletEntry.setStatus('current')
outletIndex = MibTableColumn((1, 3, 6, 1, 4, 1, 1418, 6, 5, 1, 1), Integer32().subtype(subtypeSpec=ValueRangeConstraint(0, 64))).setMaxAccess("readonly")
if mibBuilder.loadTexts: outletIndex.setStatus('current')
outletName = MibTableColumn((1, 3, 6, 1, 4, 1, 1418, 6, 5, 1, 2), DisplayString()).setMaxAccess("readwrite")
if mibBuilder.loadTexts: outletName.setStatus('current')
outletInitialState = MibTableColumn((1, 3, 6, 1, 4, 1, 1418, 6, 5, 1, 3), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(0, 1, 2))).clone(namedValues=NamedValues(("off", 0), ("on", 1), ("last", 2)))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: outletInitialState.setStatus('current')
outletCycleTime = MibTableColumn((1, 3, 6, 1, 4, 1, 1418, 6, 5, 1, 4), Integer32()).setMaxAccess("readwrite")
if mibBuilder.loadTexts: outletCycleTime.setStatus('current')
outletControl = MibTableColumn((1, 3, 6, 1, 4, 1, 1418, 6, 5, 1, 5), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(0, 1, 2))).clone(namedValues=NamedValues(("off", 0), ("on", 1), ("cycle", 2)))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: outletControl.setStatus('current')
outletStatus = MibTableColumn((1, 3, 6, 1, 4, 1, 1418, 6, 5, 1, 6), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(0, 1, 2, 3, 4))).clone(namedValues=NamedValues(("off", 0), ("on", 1), ("cycle", 2), ("reboot", 3), ("pend-on", 4)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: outletStatus.setStatus('current')
outletActualStatus = MibTableColumn((1, 3, 6, 1, 4, 1, 1418, 6, 5, 1, 7), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(0, 1))).clone(namedValues=NamedValues(("off", 0), ("on", 1)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: outletActualStatus.setStatus('current')
voltageLC1 = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 6, 1), Integer32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: voltageLC1.setStatus('current')
currentLC1 = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 6, 2), Integer32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: currentLC1.setStatus('current')
voltageLC2 = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 6, 3), Integer32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: voltageLC2.setStatus('current')
currentLC2 = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 6, 4), Integer32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: currentLC2.setStatus('current')
temperature1 = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 6, 5), Integer32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: temperature1.setStatus('current')
temperature2 = MibScalar((1, 3, 6, 1, 4, 1, 1418, 6, 6, 6), Integer32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: temperature2.setStatus('current')
outletChange = NotificationType((1, 3, 6, 1, 4, 1, 1418, 6, 7, 1)).setObjects(("IBOOTPDU-MIB", "outletIndex"), ("IBOOTPDU-MIB", "outletName"), ("IBOOTPDU-MIB", "outletStatus"))
if mibBuilder.loadTexts: outletChange.setStatus('current')
mibBuilder.exportSymbols("IBOOTPDU-MIB", deviceTemperatureUnit=deviceTemperatureUnit, sslEnabled=sslEnabled, temperature1=temperature1, serialPortEnabled=serialPortEnabled, sshEnabled=sshEnabled, device=device, ipAddress=ipAddress, snmp=snmp, snmpManagerTable=snmpManagerTable, outletEntry=outletEntry, outletControl=outletControl, snmpManagerAddress=snmpManagerAddress, webPort=webPort, rebootSystem=rebootSystem, outletTable=outletTable, network=network, deviceFamily=deviceFamily, outletInitialState=outletInitialState, outletName=outletName, macAddress=macAddress, ipMode=ipMode, deviceNumberOfOutlets=deviceNumberOfOutlets, snmpManagerEntry=snmpManagerEntry, outletActualStatus=outletActualStatus, temperature2=temperature2, webEnabled=webEnabled, sensors=sensors, dnsServer2=dnsServer2, currentLC1=currentLC1, consoleTimeout=consoleTimeout, cloudActivationCode=cloudActivationCode, gateway=gateway, sslPort=sslPort, snmpManagerName=snmpManagerName, firmwareVersion=firmwareVersion, cloudServerAddress=cloudServerAddress, telnetEnabled=telnetEnabled, outletCycleTime=outletCycleTime, outletIndex=outletIndex, console=console, TC1=TC1, setFactoryDefaults=setFactoryDefaults, deviceName=deviceName, dataprobe=dataprobe, deviceCalibrated=deviceCalibrated, iBoot_PDU_Agent=iBoot_PDU_Agent, rebootRequired=rebootRequired, sshPort=sshPort, outletStatus=outletStatus, snmpReadCommunity=snmpReadCommunity, PYSNMP_MODULE_ID=iBoot_PDU_Agent, subnetMask=subnetMask, telnetPort=telnetPort, snmpEnabled=snmpEnabled, modemCountryCode=modemCountryCode, snmpTrapCommunity=snmpTrapCommunity, deviceModelName=deviceModelName, snmpWriteCommunity=snmpWriteCommunity, deviceTimeZone=deviceTimeZone, dnsServer1=dnsServer1, voltageLC1=voltageLC1, notifications=notifications, cloudUUID=cloudUUID, deviceMaxCurrent=deviceMaxCurrent, voltageLC2=voltageLC2, snmpManagerIndex=snmpManagerIndex, outletDelayTime=outletDelayTime, cloudServiceEnabled=cloudServiceEnabled, deviceConnector=deviceConnector, outletChange=outletChange, deviceNumberOfLineCords=deviceNumberOfLineCords, currentLC2=currentLC2)



# Digital IO
IO_INFO = [
    # Right actuator: Inside the forebaffle
    {'name': 'LSR1', 'io': 'IN[1]', 'label': 'Right1Inside'},
    # Right actuator: Outside the forebaffle
    {'name': 'LSR2', 'io': 'IN[2]', 'label': 'Right2Outside'},
    # Right actuator: Spare
    {'name': 'LSR3', 'io': 'IN[3]', 'label': 'Right3NotConnected'},
    # Right actuator: Spare
    {'name': 'LSR4', 'io': 'IN[4]', 'label': 'Right4NotConnected'},
    # Left actuator: Inside the forebaffle
    {'name': 'LSL1', 'io': 'IN[5]', 'label': 'Left1Inside'},
    # Left actuator: Outside the forebaffle
    {'name': 'LSL2', 'io': 'IN[6]', 'label': 'Left2Outside'},
    # Left actuator: Spare
    {'name': 'LSL3', 'io': 'IN[7]', 'label': 'Left3NotConnected'},
    # Left actuator: Spare
    {'name': 'LSL4', 'io': 'IN[8]', 'label': 'Left4NotConnected'},
]



#!/usr/bin/env python
import argparse
import time

import requests
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock


class SynaccessAgent:
    """
    Agent to control and monitor Synaccess Networks PDUs.

    Args:
        ip_address(str): IP address for the device.
        username(str): Username credential to login to device.
        password(str): Password credential to login to device.
    """

    def __init__(self, agent, ip_address, username, password):
        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()
        self.ip_address = ip_address
        self.user = username
        self.passw = password

        agg_params = {'frame_length': 60}
        self.agent.register_feed(
            'synaccess', record=True, agg_params=agg_params)

    def __get_status(self):
        req = "http://" + self.user + ":" + self.passw + "@" +\
            self.ip_address + "/cmd.cgi?$A5"
        r = requests.get(req)
        resp = r.content.decode()[:-6:-1]  # get last 5 elements, in reverse
        return resp

    @ocs_agent.param('_')
    def get_status(self, session, params=None):
        """get_status()

        **Task** - Get the status of all outlets.

        """
        with self.lock.acquire_timeout(3, job='get_status') as acquired:
            if acquired:
                resp = self.__get_status()
                ret_str = []
                for x in resp:
                    if x == '1':
                        ret_str.append('on')
                    else:
                        ret_str.append('off')
                return True, 'status outlet [1,2,3,4,5] is {}'.format(ret_str)
            else:
                return False, "Could not acquire lock"

    @ocs_agent.param('outlet', type=int)
    def reboot(self, session, params=None):
        """reboot(outlet)

        **Task** - Reboot a given outlet.

        Parameters:
            outlet (int): The outlet that we are changing the state of.

        """
        with self.lock.acquire_timeout(3, job='reboot') as acquired:
            if acquired:
                req = "http://" + self.user + ":" + \
                    self.passw + "@" + self.ip_address + \
                    "/cmd.cgi?$A4" + " " + str(params['outlet'])
                requests.get(req)
                return True, 'Rebooted outlet {}'.format(params['outlet'])
            else:
                return False, "Could not acquire lock"

    @ocs_agent.param('outlet', type=int)
    @ocs_agent.param('on', type=bool)
    def set_outlet(self, session, params=None):
        """set_outlet(outlet, on)

        **Task** - Set a particular outlet on/off.

        Parameters:
            outlet (int): The outlet that we are changing the state of.
            on (bool): The new state. True for on, False for off.

        """
        with self.lock.acquire_timeout(3, job='set_outlet') as acquired:
            if acquired:
                if params['on']:
                    on = "1"
                else:
                    on = "0"
                req = "http://" + self.user + ":" + self.passw + "@" + \
                    self.ip_address + "/cmd.cgi?$A3" + " " + \
                    str(params['outlet']) + " " + on
                requests.get(req)
                return True, 'Set outlet {} to {}'.\
                    format(params['outlet'], params['on'])
            else:
                return False, "Could not acquire lock"

    @ocs_agent.param('on', type=bool)
    def set_all(self, session, params=None):
        """set_all(on)

        **Task** - Set all outlets on/off.

        Parameters:
            on (bool): The new state. True for on, False for off.

        """
        with self.lock.acquire_timeout(3, job='set_all') as acquired:
            if acquired:
                on = "0"
                if params['on']:
                    on = "1"
                req = "http://" + self.user + ":" + self.passw + "@" +\
                    self.ip_address + "/cmd.cgi?$A7" + " " + on
                requests.get(req)
                return True, 'Set all outlets to {}'.format(params['on'])
            else:
                return False, "Could not acquire lock"

    @ocs_agent.param('_')
    def status_acq(self, session, params=None):
        """status_acq()

        **Process** - Start data acquisition.

        Notes:
            The most recent data collected is stored in session.data in the
            structure::

                >>> response.session['data']
                {"fields":
                    {synaccess:
                        {0: 0 or 1, (0: OFF, 1:ON)
                         1: 0 or 1, (0: OFF, 1:ON)
                         2: 0 or 1, (0: OFF, 1:ON)
                         3: 0 or 1, (0: OFF, 1:ON)
                         4: 0 or 1, (0: OFF, 1:ON)
                        }
                    }
                }

        """

        with self.lock.acquire_timeout(timeout=3, job='status_acq')\
                as acquired:
            if not acquired:
                self.log.warn(
                    'Could not start status acq because {} is already running'
                    .format(self.lock.job))
                return False, 'Could not acquire lock'

            session.set_status('running')

            self.take_data = True
            last_release = time.time()
            session.data = {'fields': {}}
            while self.take_data:
                if time.time() - last_release > 1.:
                    last_release = time.time()
                    if not self.lock.release_and_acquire(timeout=10):
                        self.log.warn(
                            'Could not re-acquire lock now held by {}.'
                            .format(self.lock.job))
                        return False, 'Could not re-acquire lock (timeout)'

                current_time = time.time()
                data = {'timestamp': current_time,
                        'block_name': 'synaccess_status',
                        'data': {}}

                status_dict = {}
                resp = self.__get_status()
                for i, x in enumerate(resp):
                    if x == '1':
                        status = 1
                    else:
                        status = 0
                    data['data']['synaccess_%d' % i] = status
                    status_dict['%d' % i] = status
                self.agent.publish_to_feed('synaccess', data)
                field_dict = {'synaccess': status_dict}
                session.data['timestamp'] = current_time
                session.data['fields'] = field_dict

                time.sleep(1)  # DAQ interval
                # End of while loop

        self.agent.feeds['synaccess'].flush_buffer()
        return True, 'Acqusition exited cleanly'

    def stop_status_acq(self, session, params=None):
        if self.take_data:
            self.take_data = False
            return True, 'requested to stop taking data'

        return False, 'acq is not currently running'


def make_parser(parser=None):
    """Build the argument parser for the Agent. Allows sphinx to automatically
    build documentation based on this function.

    """
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent.
    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--ip-address', help='IP address for the device.')
    pgroup.add_argument('--username', help='Username credential to login to device.')
    pgroup.add_argument('--password', help='Password credential to login to device.')
    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args(agent_class='SynaccessAgent',
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)

    p = SynaccessAgent(agent,
                       ip_address=args.ip_address,
                       username=args.username,
                       password=args.password)
    agent.register_process('status_acq', p.status_acq,
                           p.stop_status_acq, startup=True)
    agent.register_task('get_status', p.get_status, startup={})
    agent.register_task('reboot', p.reboot)
    agent.register_task('set_outlet', p.set_outlet)
    agent.register_task('set_all', p.set_all)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



# Built-in python modules
import os
import time

# Specific module for actuator controller
on_rtd = os.environ.get('READTHEDOCS') == 'True'
if not on_rtd:
    import gclib

from socs.agents.wiregrid_actuator.drivers.DigitalIO import DigitalIO


class Actuator:
    """
    The Actuator class is for writing commands and reading status
    of the actuator via the Galil actuator controller.

    Args:
        ip_address(string)  : IP address of the actuator controller
        sleep(float)        : sleep time for every commands
        ls_list(list)       : limit-switch IO configurations
        st_list(list)       : stopper IO configurations
        verbose(int)        : verbosity level
    """

    def __init__(self, ip_address='192.168.1.100', sleep=0.05,
                 ls_list=[], st_list=[], verbose=0):
        self.ip_address = ip_address
        self.sleep = sleep
        self.verbose = verbose

        self.STOP = False  # This will become True in emergency stop.
        self.maxwaitloop = 10
        self.maxwaitloop_for_read = 1000

        # Actuator speed/distance setup
        self.speed_max = 2000
        self.speed_min = 0
        # scale factor [pulses/mm] mutiplied to distance [mm]
        self.distance_factor = 1000. / 5.5

        # Open communication to the controller
        self.g = None
        self._connect()

        # Initialize Digital IO classes
        # for limit-switch & stopper
        self.ls = DigitalIO(
            'limit-switch', ls_list, self.g, get_onoff_reverse=False)
        self.st = DigitalIO(
            'stopper', st_list, self.g,
            get_onoff_reverse=False, set_onoff_reverse=False)

    def __del__(self):
        self._cleanG()
        print('Actuator:__del__(): '
              'Successfully close the actuator controller.')

    ######################
    # Internal functions #
    ######################

    # Send command & get return function
    # If an error occurs, it raises an error
    def _command(self, cmd, doSleep=True):
        if self.verbose > 0:
            print('Actuator:_command(): command = "{}"\\n'.format(cmd))
        try:
            ret = self.g.GCommand(cmd)
        except Exception as e:
            msg = 'Actuator:_command(): ERROR!: Failed to send command({}).\n'\
                  'Actuator:_command(): ERROR!: Exception = "{}"'\
                  .format(cmd, e)
            print(msg)
            raise
        if self.verbose > 0:
            print('Actuator:_command(): response = "{}"\\n'.format(ret))
        if doSleep:
            time.sleep(self.sleep)
        return ret

    def _cleanG(self):
        if self.g is not None:
            self.g.GClose()
            del self.g
            self.g = None
        if self.verbose > 0:
            print('Actuator:_cleanG(): '
                  'Successfully cleaned the Galil connection')
        return True

    def _connect(self):
        self._cleanG()
        # Open communication to the controller
        print('Actuator:_connect(): '
              'Initialize the Galil actuator controller')
        self.g = gclib.py()
        if self.g is None:
            msg = 'Actuator:_connect() : ERROR!: Failed to '\
                  'initialize the connection to the actuator controller.'
            raise RuntimeError(msg)
        self.g.GOpen('{}'.format(self.ip_address))
        # Connection check
        print('Actuator:_connect(): {}'.format(self.g.GInfo()))
        status = self._check_motortype()
        if not status:
            print('Actuator:_connect(): WARNING! '
                  'Motor type is not correct!')
            print('Actuator:_connect(): '
                  '--> Power off & change the motor types!')
            # Set controller parameters
            # Motor OFF (need for MT command)
            self._set_motor_onoff(onoff=False)
            # Motor type: stepper with active low(2)/high(2.5) step pulses
            self._command('MT 2,2')
        # Motor ON (A,B,N[virtual gear])
        self._set_motor_onoff(onoff=True)
        self._set_actuator_parameters()

        status = self.check_connect()
        if not status:
            msg = 'Actuator:_connect(): ERROR!: '\
                  'check_connect() is failed.'
            raise RuntimeError(msg)
        time.sleep(1)
        print('Actuator:_connect(): Successfully make a connection.')
        return True

    # Set motor ON/OFF
    # Args: onoff = True or False
    def _set_motor_onoff(self, onoff):
        if onoff:  # ON
            self._command('SH AB')
        else:  # OFF
            self._command('MO')
        return True

    # Check the motor type ([2,2] is correct.)
    def _check_motortype(self):
        try:
            ret = self._command('MT ?,?')
            mts = [int(float(motor_type)) for motor_type in ret.split(',')]
        except Exception as e:
            msg = \
                'Actuator:_check_motortype(): ERROR!: Failed to check '\
                'the motor type! | '\
                'Exception = "{}"'.format(e)
            raise RuntimeError(msg)
        if len(mts) != 2:
            print('Actuator:_check_motortype(): '
                  'WARNING!: Returned motor type = {}. '
                  'Array size is not correct.'
                  .format(mts))
            return False
        else:
            if not (mts[0] == 2 and mts[1] == 2):
                print('Actuator:_check_motortype(): '
                      'WARNING!: Returned motor type = {}. '
                      'Motor type is not correct (should be [2,2]).'
                      .format(mts))
                return False
        if self.verbose > 0:
            print('Actuator:_check_motortype(): '
                  'Motor type is correct!')
        return True

    def _set_actuator_parameters(self):
        # Stop motion
        self._command('ST')
        # Set master axis of A & B is N
        self._command('GAA=N')
        self._command('GAB=N')
        # Set gear to N
        self._command('GRA=1')
        self._command('GRB=-1')
        # Smoothing pulse: sample value=16
        self._command('KSA=16')
        self._command('KSB=16')
        # Set current: 3A
        self._command('AGA=3')
        self._command('AGB=3')
        # Set microstepping: 1/8
        self._command('YAA=8')
        self._command('YAB=8')
        # Set motor resolutio: 200 steps/revolution = 1.8deg/step
        self._command('YBA=200')
        self._command('YBB=200')
        self._command('YBN=200')
        # Set speed: speed_max
        self._command('SPA={}'.format(self.speed_max))
        self._command('SPB={}'.format(self.speed_max))
        self._command('SPN={}'.format(self.speed_max))
        # Set positioin
        self._command('PRA=0')
        self._command('PRB=0')
        self._command('PRN=0')
        if self.verbose > 0:
            print('Actuator:_set_actuator_parameters(): '
                  'Successfully set the actuator controller parameters!')
        return True

    ##################
    # Main functions #
    ##################

    # Return: success (True) or failure (False)
    def move(self, distance, speedrate=0.1):
        self._set_actuator_parameters()
        print('Actuator:move(): distance = {}, speedrate = {}'
              .format(distance, speedrate))
        if self.STOP:
            msg = 'Actuator:move(): WARNING!: Did NOT move due to STOP flag.'
            print(msg)
            return False
        if speedrate < 0. or speedrate > 1.:
            print('Actuator:move(): WARNING!: '
                  'Speedrate should be between 0 and 1.')
            print('Actuator:move(): WARNING!: '
                  'Speedrate is sed to 0.1.')
            speedrate = 0.1
        speed = \
            int(speedrate * (self.speed_max - self.speed_min) + self.speed_min)
        # distance_count is an absolute value
        distance_count = int(abs(distance) * self.distance_factor)
        # Change the gearing
        if distance >= 0.:
            # Forwarding
            self._command('GRA=1')
            self._command('GRB=-1')
            dlabel = 'forwarding'
        else:
            # Backwarding
            self._command('GRA=-1')
            self._command('GRB=1')
            dlabel = 'backwarding'
        if self.verbose > 0:
            print('Actuator:move(): distance_count = {} ({})'
                  .format(distance_count, dlabel))
        # Set the speed and distance
        self._command('SPA={}'.format(0))
        self._command('SPB={}'.format(0))
        self._command('SPN={}'.format(speed))
        self._command('PRA={}'.format(0))
        self._command('PRB={}'.format(0))
        self._command('PRN={}'.format(distance_count))
        # Start motion
        print('Actuator:move(): Start the moving...')
        self._command('BGN')
        if self.verbose > 0:
            print('Actuator:move(): Succsessfully sent the moving commands!')
        return True

    # Return 1 (running) or 0 (stopping)
    def is_run(self):
        ret = self._command('MG _BGN', doSleep=True)
        isrun = (int)((float)(ret))
        if self.verbose > 0:
            print('Actuator:is_run() : running status = "{}"'.format(isrun))
        return isrun

    # Wait for the end of moving
    # Args: max_loop_time = maximum waiting time [sec]
    # Return: success (True) or failure (False)
    def wait_idle(self, max_loop_time=180):
        # Number of loop for max_loop_time [sec]
        max_loop = int(max_loop_time / self.sleep)
        for i in range(max_loop):
            isrun = self.is_run()
            if not isrun:
                print('Actuator:wait_idle(): The running is finished!')
                return True
        print('Actuator:wait_idle(): ERROR!: '
              'Exceed max. number of loop ({} times)'.format(i))
        return False

    # Check the connection
    def check_connect(self):
        status = self._check_motortype()
        if not status:
            print('Actuator:check_connect(): ERROR!: '
                  'The connection to the actuator controller is BAD!')
            return False
        else:
            if self.verbose > 0:
                print('Actuator:check_connect(): '
                      'The connection to the actuator controller is OK!')
            return True

    def reconnect(self):
        print('Actuator:reconnect() : *** Trying to reconnect... ***')
        for i in range(self.maxwaitloop):
            # reconnect
            print('Actuator:reconnect(): {}th try to reconnection'.format(i))
            try:
                self._connect()
            except Exception as e:
                msg = 'Actuator:reconnect(): WARNING!: '\
                      'Failed to reconnect to the actuator controller! (i={})'\
                      ' | Exception = "{}"'.format(i, e)
                print(msg)
                time.sleep(1)
                continue
            # Set g on DigitalIOs
            self.ls.g = self.g
            self.st.g = self.g
            print('Actuator:reconnect(): '
                  'Successfully reconnect to the actuator controller!')
            return True
        print('Actuator:reconnect(): ERROR!: Exceed the max. number of '
              'trying to reconnect to the actuator controller.')
        return False

    # Get motor ON/OFF
    # Return: 1 (ON) or 0 (OFF)
    def get_motor_onoff(self):
        try:
            # 0: ON, 1:OFF
            retA = self._command('MG _MOA')
            retB = self._command('MG _MOB')
            onoffA = int(float(retA))
            onoffB = int(float(retB))
            # invert to 0: OFF, 1: ON
            onoffA = int(not onoffA)
            onoffB = int(not onoffB)
        except Exception as e:
            msg =\
                'Actuator:get_motor_onoff(): ERROR!: '\
                'Failed to get motor on/off! | '\
                'Exception = "{}"'.format(e)
            print(msg)
            raise
        if onoffA == 1 and onoffB == 1:
            return 1
        elif onoffA == 0 and onoffB == 0:
            return 0
        else:
            msg =\
                'Actuator:get_motor_onoff(): ERROR!: '\
                'Strange status of motor on/off! '\
                'motor A = {} / motor B = {}'.format(
                    'ON' if onoffA else 'OFF',
                    'ON' if onoffB else 'OFF')
            raise RuntimeError(msg)

    # Set motor ON/OFF
    # Args: onoff = True or False
    def set_motor_onoff(self, onoff):
        try:
            self._set_motor_onoff(onoff)
        except Exception as e:
            msg = \
                'Actuator:set_motor_onoff(): ERROR!: '\
                'Failed to set motor {}! | '\
                'Exception = "{}"'\
                .format('ON' if onoff else 'OFF', e)
            print(msg)
            raise
        onoff_test = self.get_motor_onoff()
        if onoff_test != int(onoff):
            msg = \
                'Actuator:set_motor_onoff(): ERROR!: '\
                'Set motor {} but the current ON/OFF ("{}") is different!'\
                .format('ON' if onoff else 'OFF', onoff_test)
            raise RuntimeError(msg)
        print('Actuator:set_motor_onoff(): '
              'Successfully {} the actuator motors!'
              .format('ON' if onoff else 'OFF'))
        return True

    # Return: success (True) or failure (False)
    def stop(self):
        if self.verbose > 0:
            print('Actuator:stop(): Stop the actuator')
        for i in range(self.maxwaitloop):
            self._command('ST')
            isrun = self.is_run()
            if not isrun:
                if self.verbose > 0:
                    print('Actuator:stop(): Successfully stop the actuator!')
                return True
            print('Actuator:stop(): WARNING!: '
                  'Could not stop the actuator! --> Retry')
        print('Actuator:stop(): ERROR!: '
              'Exceed the max. number of retries ({} times).'.format(i))
        return False

    # Return: success (True) or failure (False)
    def hold(self):
        if self.verbose > 0:
            print('Actuator:hold(): Hold the actuator')
        self.STOP = True
        ret = self.stop()
        return ret

    # Release(unhold) the hold state
    def release(self):
        if self.verbose > 0:
            print('Actuator:release(): Release the actuator from hold state')
        self.STOP = False
        if self.verbose > 0:
            print('Actuator:release(): Successfully release the actuator!')
        return True



import argparse
import csv
import os
import struct
import time

import numexpr
import numpy as np
import txaio
import yaml
from labjack import ljm
from labjack.ljm.ljm import LJMError
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock
from scipy.interpolate import interp1d

txaio.use_twisted()


# Convert Data
def float2int(num):
    return struct.unpack("=i", struct.pack("=f", num))[0]


def concatData(data):
    tVal = 0
    upper = True
    for reg in data:
        if upper:
            tVal = ((reg & 0xFFFF) << 16)
            upper = False
        else:
            tVal = tVal | (reg & 0xFFFF)
            upper = True
    return tVal


# Converting numbers to 16-bit data arrays
def uint16_to_data(num):
    return struct.unpack("=H", struct.pack("=H", num & 0xFFFF))[0]


def uint32_to_data(num):
    data = [0, 0]
    data[0] = struct.unpack("=H", struct.pack("=H", (num >> 16) & 0xffff))[0]
    data[1] = struct.unpack("=H", struct.pack("=H", num & 0xffff))[0]
    return data


def int32_to_data(num):
    data = [0, 0]
    data[0] = struct.unpack("=H", struct.pack("=H", (num >> 16) & 0xffff))[0]
    data[1] = struct.unpack("=H", struct.pack("=H", num & 0xffff))[0]
    return data


def float32_to_data(num):
    intNum = float2int(num)
    data = [0, 0]
    data[0] = (intNum >> 16) & 0xFFFF
    data[1] = intNum & 0xFFFF
    return data


# Converting data arrays to numbers
def data_to_uint16(data):
    return data[0]


def data_to_uint32(data):
    return concatData(data)


def data_to_int32(data):
    return struct.unpack("=i", struct.pack("=I", concatData(data)))[0]


def data_to_float32(data):
    return struct.unpack("=f", struct.pack("=I", concatData(data)))[0]


class LabJackFunctions:
    """
    Labjack helper class to provide unit conversion from analog input voltage
    """

    def __init__(self):
        self.log = txaio.make_logger()

    def unit_conversion(self, v_array, function_info):
        """
        Given a voltage array and function information from the
        labjack_config.yaml file, applies a unit conversion.
        Returns the converted value and its units.

        Args:
            v_array (np.array):
                The voltages to be converted.
            function_info (dict):
                Specifies the type of function. If custom, also gives the function.
        """
        if function_info["user_defined"] == 'False':
            function = getattr(self, function_info['type'])
            return function(v_array)

        # Custom function evaluation
        else:
            units = function_info['units']
            new_values = []
            for v in v_array:
                new_values.append(float(numexpr.evaluate(function_info["function"])))
            return new_values, units

    def MKS390(self, v_array):
        """
        Conversion function for the MKS390 Micro-Ion ATM
        Modular Vaccum Gauge.
        """
        value = 1.3332 * 10**(2 * v_array - 11)
        units = 'mBar'
        return value, units

    def warm_therm(self, v_array):
        """
        Conversion function for SO warm thermometry readout.
        Voltage is converted to resistance using the LJTick, which
        has a 2.5V supply and 10kOhm reference resistor. Resistance
        is converted to degrees Celsius using the calibration curve
        for the thermistor model, serial number 10K4D25.
        """
        # LJTick voltage to resistance conversion
        R = (2.5 - v_array) * 10000 / v_array

        # Import the Ohms to Celsius cal curve and apply cubic
        # interpolation to find the temperature
        reader = csv.reader(open('cal_curves/GA10K4D25_cal_curve.txt'),
                            delimiter=' ')
        lists = [el for el in [row for row in reader]]
        T_cal = np.array([float(RT[0]) for RT in lists[1:]])
        R_cal = np.array([float(RT[1]) for RT in lists[1:]])
        T_cal = np.flip(T_cal)
        R_cal = np.flip(R_cal)
        try:
            RtoT = interp1d(R_cal, T_cal, kind='cubic')
            values = RtoT(R)

        except ValueError:
            self.log.error('Temperature outside thermometer range')
            values = -1000 + np.zeros(len(R))

        units = 'C'

        return values, units


class LabJackAgent:
    """Agent to collect data from LabJack device.

    Parameters:
        agent (OCSAgent): OCSAgent object for the Agent.
        ip_address (str): IP Address for the LabJack device.
        active_channels (str or list): Active channel description, i.e.
            'T7-all', 'T4-all', or list of channels in form ['AIN0', 'AIN1'].
        function_file (str): Path to file for unit conversion.
        sampling_frequency (float): Sampling rate in Hz.

    """

    def __init__(self, agent, ip_address, active_channels, function_file,
                 sampling_frequency):
        self.active = True
        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()
        self.ip_address = ip_address
        self.module = None
        self.ljf = LabJackFunctions()
        self.sampling_frequency = sampling_frequency

        # Labjack channels to read
        if active_channels[0] == 'T7-all':
            self.chs = ['AIN{}'.format(i) for i in range(14)]
        elif active_channels[0] == 'T4-all':
            self.chs = ['AIN{}'.format(i) for i in range(12)]
        else:
            self.chs = active_channels

        # Load dictionary of unit conversion functions from yaml file. Assumes
        # the file is in the $OCS_CONFIG_DIR directory
        if function_file == 'None':
            self.functions = {}
        else:
            function_file_path = os.path.join(os.environ['OCS_CONFIG_DIR'],
                                              function_file)
            with open(function_file_path, 'r') as stream:
                self.functions = yaml.safe_load(stream)
                if self.functions is None:
                    self.functions = {}
                self.log.info(f"Applying conversion functions: {self.functions}")

        self.initialized = False
        self.take_data = False

        # Register main feed. Exclude influx due to potentially high scan rate
        agg_params = {
            'frame_length': 60,
            'exclude_influx': True
        }
        self.agent.register_feed('sensors',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=1)

        # Register downsampled feed for influx.
        agg_params_downsampled = {
            'frame_length': 60
        }
        self.agent.register_feed('sensors_downsampled',
                                 record=True,
                                 agg_params=agg_params_downsampled,
                                 buffer_time=1)
        self.agent.register_feed('registers',
                                 record=True,
                                 agg_params={'frame_length': 10 * 60},
                                 buffer_time=1.)

    # Task functions
    def init_labjack(self, session, params=None):
        """init_labjack(auto_acquire=False)

        **Task** - Initialize LabJack module.

        Parameters:
            auto_acquire (bool): Automatically start acq process after
                initialization. Defaults to False.

        """
        if self.initialized:
            return True, "Already initialized module"

        with self.lock.acquire_timeout(0, job='init') as acquired:
            if not acquired:
                self.log.warn("Could not start init because "
                              "{} is already running".format(self.lock.job))
                return False, "Could not acquire lock."

            session.set_status('starting')
            # Connect with the labjack
            self.handle = ljm.openS("ANY", "ANY", self.ip_address)
            info = ljm.getHandleInfo(self.handle)
            self.log.info("\nOpened LabJack of type: %i, Connection type: %i,\n"
                          "Serial number: %i, IP address: %s, Port: %i" %
                          (info[0], info[1], info[2],
                           ljm.numberToIP(info[3]), info[4]))

        session.add_message("Labjack initialized")

        self.initialized = True

        # Start data acquisition if requested in site-config
        auto_acquire = params.get('auto_acquire', False)
        auto_acquire_reg = params.get('auto_acquire_reg', False)

        if auto_acquire:
            self.agent.start('acq')
        if auto_acquire_reg:
            self.agent.start('acq_reg')
        return True, 'LabJack module initialized.'

    def acq(self, session, params=None):
        """acq(sampling_freq=2.5)

        **Process** - Acquire data from the Labjack.

        Parameters:
            sampling_frequency (float):
                Sampling frequency for data collection. Defaults to 2.5 Hz.

        """
        if params is None:
            params = {}

        # Setup streaming parameters. Data is collected and published in
        # blocks at 1 Hz or the scan rate, whichever is less.
        scan_rate_input = params.get('sampling_frequency',
                                     self.sampling_frequency)
        scans_per_read = max(1, int(scan_rate_input))
        num_chs = len(self.chs)
        ch_addrs = ljm.namesToAddresses(num_chs, self.chs)[0]

        with self.lock.acquire_timeout(0, job='acq') as acquired:
            if not acquired:
                self.log.warn("Could not start acq because "
                              "{} is already running".format(self.lock.job))
                return False, "Could not acquire lock."

            session.set_status('running')
            self.take_data = True

            # Start the data stream. Use the scan rate returned by the stream,
            # which should be the same as the input scan rate.
            try:
                scan_rate = ljm.eStreamStart(self.handle, scans_per_read, num_chs,
                                             ch_addrs, scan_rate_input)
            except LJMError as e:  # in case the stream is running
                self.log.error(e)
                self.log.error("Stopping previous stream and starting new one")
                ljm.eStreamStop(self.handle)
                scan_rate = ljm.eStreamStart(self.handle, scans_per_read, num_chs,
                                             ch_addrs, scan_rate_input)
            self.log.info(f"\nStream started with a scan rate of {scan_rate} Hz.")

            cur_time = time.time()
            while self.take_data:
                data = {
                    'block_name': 'sens',
                    'data': {}
                }

                # Query the labjack
                raw_output = ljm.eStreamRead(self.handle)
                output = raw_output[0]

                # Data comes in form ['AIN0_1', 'AIN1_1', 'AIN0_2', ...]
                for i, ch in enumerate(self.chs):
                    ch_output = output[i::num_chs]
                    data['data'][ch + 'V'] = ch_output

                    # Apply unit conversion function for this channel
                    if ch in self.functions.keys():
                        new_ch_output, units = \
                            self.ljf.unit_conversion(np.array(ch_output),
                                                     self.functions[ch])
                        data['data'][ch + units] = list(new_ch_output)

                # The labjack outputs at exactly the scan rate but doesn't
                # generate timestamps. So create them here.
                timestamps = [cur_time + i / scan_rate for i in range(scans_per_read)]
                cur_time += scans_per_read / scan_rate
                data['timestamps'] = timestamps

                self.agent.publish_to_feed('sensors', data)

                # Publish to the downsampled data feed only the first
                # timestamp and data point for each channel.
                data_downsampled = {
                    'block_name': 'sens',
                    'data': {},
                    'timestamp': timestamps[0]
                }
                for key, value in data['data'].items():
                    data_downsampled['data'][key] = value[0]
                self.agent.publish_to_feed('sensors_downsampled', data_downsampled)
                session.data = data_downsampled

            # Flush buffer and stop the data stream
            self.agent.feeds['sensors'].flush_buffer()
            self.agent.feeds['sensors_downsampled'].flush_buffer()
            ljm.eStreamStop(self.handle)
            self.log.info("Data stream stopped")

        return True, 'Acquisition exited cleanly.'

    def _stop_acq(self, session, params=None):
        if self.take_data:
            self.take_data = False
            return True, 'requested to stop taking data.'
        else:
            return False, 'acq is not currently running'

    def acq_reg(self, session, params=None):
        """acq_reg(sampling_frequency=2.5)

        **Task** - Start data acquisition when you want to read out
        non-standard registers.

        In particular the custom registers labjack was built for reading out
        thermocouples. Maximum is about 2.5 Hz but is set by the register read
        time which is estimated at the beginning of the acq_reg setup step.

        Args:
            sampling_frequency (float):
                Sampling frequency for data collection. Defaults to 2.5 Hz
                Maximum set by the register read time. Will reset to lower
                rate if faster than possible read time.
        """
        if params is None:
            params = {}
        # Determine the read time latency to set the max allowable
        # sampling rate by reading the register 100 times in a row
        # and recording the time it takes to read each time. Then
        # setting the max sample rate to be 50mS greater than the median
        # of the time it took to read.
        num_chs = len(self.chs)
        times = []
        for i in range(100):
            times.append(time.time())
            ljm.eReadNames(self.handle, num_chs, self.chs)[0]
        read_dt = np.round(np.median(np.diff(times)), 2) + 0.05
        max_fsamp = min(2.5, 1 / read_dt)
        # Setup streaming parameters. Data is collected and published in
        # blocks at 1 Hz or the scan rate, whichever is less.
        scan_rate_input = params.get('sampling_frequency',
                                     self.sampling_frequency)
        # Warn user that they input too fast of a sample rate and set
        # to maximum allowable.
        if scan_rate_input > max_fsamp:
            self.log.warn(f'Sampling rate {scan_rate_input} exceeds'
                          'allowable range for register read mode'
                          'if you want to sample faster please use'
                          'standard acquire mode and do not read out '
                          'any non-standard channels. Setting sampling'
                          f'rate to maximum = {max_fsamp} Hz')
            scan_rate_input = max_fsamp
            scan_rate_dt = 0
        else:
            scan_rate_dt = (1 / scan_rate_input) - read_dt

        with self.lock.acquire_timeout(0, job='acq_reg') as acquired:
            if not acquired:
                self.log.warn("Could not start acq because "
                              "{} is already running".format(self.lock.job))
                return False, "Could not acquire lock."

            session.set_status('running')
            self.take_data = True

            while self.take_data:
                data = {
                    'block_name': 'reg',
                    'data': {}
                }

                # Get a timestamp
                timestamp = time.time()
                # Query the labjack
                output = ljm.eReadNames(self.handle, num_chs, self.chs)

                # Data comes in form ['reg1', 'reg2', 'reg3', ...]
                # where regn are the registers in your active_channels
                # list in your site config file.
                for i, ch in enumerate(self.chs):
                    ch_output = output[i]
                    data['data'][ch] = ch_output

                    # Apply unit conversion function for this channel
                    if ch in self.functions.keys():
                        new_ch_output, units = \
                            self.ljf.unit_conversion(ch_output,
                                                     self.functions[ch])
                        data['data'][ch + units] = new_ch_output

                data['timestamp'] = timestamp
                self.agent.publish_to_feed('registers', data)
                time.sleep(scan_rate_dt)

            # Flush buffer and stop the data stream
            self.agent.feeds['registers'].flush_buffer()
            ljm.close(self.handle)
            self.log.info("Data stream stopped")
        return True, 'Acquisition exited cleanly.'


def make_parser(parser=None):
    """
    Build the argument parser for the Agent. Allows sphinx to automatically
    build documentation based on this function.
    """
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent.
    pgroup = parser.add_argument_group('Agent Options')

    ip_txt = 'ip-address of LabJack'
    pgroup.add_argument('--ip-address', help=ip_txt)
    achan_txt = 'Channels or register names to readout default is to read out all '
    achan_txt += 'available outputs. But typically you would want to pass a list of '
    achan_txt += 'analog inputs (i.e. ["AIN0", "AIN1", "AIN2"]) or a list of custom '
    achan_txt += 'registers if using ``--mode=acq_reg``.'
    pgroup.add_argument('--active-channels',
                        default=['T7-all'], nargs='+', help=achan_txt)
    ffile_txt = 'Path to file that defines functions for converting analog inputs to '
    ffile_txt += 'useful units (i.e. temp, pressure, etc.)'
    pgroup.add_argument('--function-file', default='None', help=ffile_txt)
    fs_txt = 'This is the rate each channel in the `active-channels` list is sampled '
    fs_txt += 'at. In ``--mode=acq_reg`` the maximum for this is 2.5 Hz can sample much '
    fs_txt += 'faster O(kHz) in ``--mode=acq``.'
    pgroup.add_argument('--sampling-frequency', default='2.5', help=fs_txt)
    acq_txt = 'Options are **acq**: read out analog inputs (can stream quickly)'
    acq_txt += ', **acq_reg**: read out custom configured registers'
    acq_txt += ', or **idle**: leave device idle at startup.'
    pgroup.add_argument('--mode', default='acq',
                        choices=['idel', 'acq', 'acq_reg'], help=acq_txt)

    return parser


def main(args=None):
    # Start logging
    txaio.start_logging(level=os.environ.get("LOGLEVEL", "info"))

    parser = make_parser()
    args = site_config.parse_args(agent_class='LabJackAgent',
                                  parser=parser,
                                  args=args)

    init_params = False
    if args.mode == 'acq':
        init_params = {'auto_acquire': True}
    if args.mode == 'acq_reg':
        init_params = {'auto_acquire_reg': True}

    ip_address = str(args.ip_address)
    active_channels = args.active_channels
    function_file = str(args.function_file)
    sampling_frequency = float(args.sampling_frequency)

    agent, runner = ocs_agent.init_site_agent(args)

    sensors = LabJackAgent(agent,
                           ip_address=ip_address,
                           active_channels=active_channels,
                           function_file=function_file,
                           sampling_frequency=sampling_frequency)

    agent.register_task('init_labjack',
                        sensors.init_labjack,
                        startup=init_params)
    agent.register_process('acq', sensors.acq, sensors._stop_acq)
    agent.register_process('acq_reg', sensors.acq_reg,
                           sensors._stop_acq)
    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()
